/*
	Blokus SDL
		blmisc.cpp

	This file includes some video handling funtions
	for class APP.

	Copyright (c) 2003-2004 Antti Partanen


	This file is part of Blokus SDL.

	Blokus SDL is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Blokus SDL is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with Blokus SDL; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#define ERROR_STRINGS_INUSE

//**********************************************************
//* INCLUDES
#include "dmf.h"
#include "derrors.h"
#include "txtfile.h"

#include "blapp.h"
#include "blvideo.h"
#include "blversion.h"
#include "blmisc.h"
#include "blnick.h"
#include "bljoin.h"
#include "blfiles.h"
#include "blokus.h"
#include "blexceptions.h"

//**********************************************************
//* METHODS

//**********************************************************
//* class APP

//**********************************************************
/**
	Change or set video mode.
*/
EVALUE APP::change_video_mode(void)
{
	/* Variables. */
	static bool reset = false;
	EVALUE err;
	int enable_unicode;

	/*
		If the video is already initialized,
		then it must be shutdown and reinit.
	*/
	if (reset)
	{
		/* Get unicode translation state.*/
		enable_unicode = SDL_EnableUNICODE(0);

		graph_release();
		SDL_QuitSubSystem(SDL_INIT_VIDEO);
		SDL_InitSubSystem(SDL_INIT_VIDEO);
	}

	/*
		Find out is the video mode,
		which is going to be set, OK?
	*/
	Uint32 bpp;

	switch (st.bpp)
	{
	case 0:
		bpp = 8;
		break;

	case 1:
		bpp = 16;
		break;

	case 2:
		bpp = 24;
		break;

	default:
	case 3:
		bpp = 32;
		break;
	}

	/* Force resolution to 640x480. */
	if (st.w != 640)
	{
		st.w = 640;
	}

	if (st.h != 480)
	{
		st.h = 480;
	}

	/* If aalib compilation. */
	if (st.fullscreen != 0)
	{
		err = SDL_VideoModeOK(st.w, st.h, bpp, SDL_HWSURFACE | SDL_DOUBLEBUF | SDL_FULLSCREEN);
	}
	else
	{
		err = SDL_VideoModeOK(st.w, st.h, bpp, SDL_HWSURFACE | SDL_DOUBLEBUF);
	}
	
	if (err != 16 && err != 32 && err != 24 && err != 8)
	{
		FILE_Text(TXTFILE_INFO, true, false, false,
			      "ERROR: Video mode unavailable: %ix%ix%i.\n"
				  , st.w, st.h, bpp);
		return (E_FAILED);
    }
	else if (err != bpp)
	{
		FILE_TextI("WARNING: Video mode unavailable: %ix%ix%i.\n"
				  , st.w, st.h, bpp);
		FILE_TextI("WARNING: Using instead mode: %ix%ix%i.\n"
				  , st.w, st.h, err);
		bpp = err;
    }

	switch (bpp)
	{
	case 8:
		st.bpp = 1;
		break;

	case 16:
		st.bpp = 1;
		break;

	case 24:
		st.bpp = 2;
		break;

	default:
	case 32:
		st.bpp = 3;
		break;
	}

	if (st.fullscreen != 0)
	{
		err = VID_Init(st.w, st.h, bpp, SDL_HWSURFACE | SDL_DOUBLEBUF | SDL_FULLSCREEN);
	}
	else
	{
		err = VID_Init(st.w, st.h, bpp, SDL_HWSURFACE | SDL_DOUBLEBUF);
	}
	
	vscreen = SDL_GetVideoSurface();

	if (err != A_OK || !vscreen)
	{
		FILE_Text(TXTFILE_INFO, true, false, false, 
			      "ERROR: Unable to init video mode %ix%ix%i"
			      ", Init() returned: %s, SDL error: %s.\n"
				  , vscreen->w, vscreen->h, vscreen->format->BitsPerPixel,
				  ERROR_STRINGS[err], SDL_GetError());
		return (E_FAILED);
	}

	//* TODO: fix this function.. load palette rather from DMF
/*	if (bpp == 8)
	{
		fstream pal;

		pal.open("graph/palette.act", ios::in | ios::binary | ios::nocreate);

		if (pal.fail())
		{
			FILE_Text(TXTFILE_INFO, true, false, false, 
				"ERROR: Unable to load palette: \"graph/palette.act\"\n");
			return (E_FAILED);
		}

		char buffer[256 * 3];

		pal.read(buffer, 256 * 3);
		pal.close();

		int i;
		int buffer_pntr = 0;
		SDL_Color colors[256];
		
		for (i = 0; i < 256; i++)
		{
			colors[i].r = buffer[buffer_pntr];
			colors[i].g = buffer[buffer_pntr + 1];
			colors[i].b = buffer[buffer_pntr + 2];

			buffer_pntr += 3;
		}

		SDL_SetPalette(vscreen, SDL_LOGPAL | SDL_PHYSPAL, colors, 0, 256);
	}*/

	/* Print some extra information. */
	FILE_TextE("EXTRA: Asked video mode: %ix%ix%i, got video mode: %ix%ix%i.\n",
			  st.w,
			  st.h,
			  bpp,
			  vscreen->w,
			  vscreen->h,
			  vscreen->format->BitsPerPixel);

	/* Becose of reset we must reload graphics... */
	if (reset)
	{
		reload_bg = true;
		
		err = graph_load();

		if (err != A_OK)
		{
			return (err);
		}

		/* Set keyboard repeat. */
		if (SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY,
			SDL_DEFAULT_REPEAT_INTERVAL) != 0)
		{
			FILE_TextI("WARNING: Strange, failed enabling keyboard repeat...\n");
		}

		/* Set unicode translation. */
		SDL_EnableUNICODE(enable_unicode);
	}

	reset = true;

	return (A_OK);
};

//**********************************************************
/**
	Load some graphics.
*/
EVALUE APP::graph_load(void)
{
	FILE_TextD("DEBUG: Entered LoadGraph()\n");
   
	/* General. */
	SDL_WM_SetCaption(ID_STRING, ID_STRING);

	/* Mouse. */
	SDL_ShowCursor(SDL_DISABLE);
	FILE_TextE("EXTRA DEBUG: Loading mouse cursor...\n");
	LoadDMF(&Cursor, MOUSE, DMF_A_NAME, true);

	/* For debug. */
	FILE_TextD("DEBUG: Loading some fonts...\n");
	
	/* iFont, used for tooltips and some others. */
	FILE_TextD("DEBUG: Loading iFont (id: %u)...\n", FONT_I_NAME);
        iFont = load_ttf(FONT_I_NAME, FONT_I_SIZE, DMF_A_NAME);
	TTF_SetFontStyle(iFont , FONT_I_STYLE);
	if (!iFont)
	{
		FILE_Text(TXTFILE_INFO, true, false, false,
			"ERROR: Couldn't load font (id: %u).\n", FONT_I_NAME);
		return (E_NULL);
	}

	/* Font for menu buttons. */
	FILE_TextD("DEBUG: Loading MFont (id: %u)...\n", MENU_FONT_NAME);
	MFont = load_ttf(MENU_FONT_NAME, MENU_FONT_SIZE, DMF_A_NAME);
	TTF_SetFontStyle(MFont, MENU_FONT_STYLE);
	if (!MFont)
	{
		FILE_Text(TXTFILE_INFO, true, false, false,
			"ERROR: Couldn't load MFont (id: %u).\n", MENU_FONT_NAME);
		return (E_NULL);
	}

	FILE_TextD("DEBUG: Some fonts loaded.\n");

	/* Setup Id_string and Version surfaces. */
	FILE_TextD("DEBUG: Setup Id_string and Version surfaces.\n");
	Id_string = VID_RenderTTF(ID_STRING, &LText, iFont);
	Version = VID_RenderTTF(VERSION_STRING, &LText, iFont);

	/* Menu list foreground graphics. */
	LoadDMF(&foreMenuList, FORELIST, DMF_A_NAME, true, 128);

	/* Load mode specific graphics, mostly fonts. */
	graph_load_by_mode(Mode);

	/* Some debugging information. */
	FILE_TextD("DEBUG: exiting LoadGraph() without errors...\n");
   
	/* Return. */
	return (A_OK);
}

//**********************************************************
/**
	Release reserved graphic resources.
*/
void APP::graph_release(void)
{
	Cursor.~DPIC();
	BG.~DPIC();

	if (Id_string)
	{
		SDL_FreeSurface(Id_string);
	}

	if (Version)
	{
		SDL_FreeSurface(Version);
	}

	if (dFont)
	{
		TTF_CloseFont(dFont);
		dFont = NULL;
	}

	if (fChat)
	{
		TTF_CloseFont(fChat);
		fChat = NULL;
	}

	if (iFont)
	{
		TTF_CloseFont(iFont);
		iFont = NULL;
	}

	if (MFont)
	{
		TTF_CloseFont(MFont);
		MFont = NULL;
	}
}

//**********************************************************
/**
	Load given TTF-font from DMF.
*/
TTF_Font *APP::load_ttf(unsigned short id, int size, char *dmf)
{
	DMFINFO dtfinfo;
	DMFHEAD dtfhead;

	if (DMF_ReadHead(&dtfhead, dmf) != 0)
	{
		FILE_TextV("ERROR: DMF_ReadHead() failed to open DTF: %s\n",
			       dmf);
		return (NULL);
	}
 
	if (DMF_ReadInfo(id, &dtfinfo, dmf) != A_OK)
	{
		FILE_TextV("ERROR: DMF_ReadInfo() failed to read"
			       " info for picture id %d\n", id);
		return (NULL);
	}
	
	int data_size;
	char *data = DMF_DataRead(id, &data_size, dmf);

	if (!data)
	{
		return (NULL);
	}

	SDL_RWops *rw_ops = SDL_RWFromMem(data, data_size);

	TTF_Font *font;

	font = TTF_OpenFontIndexRW(rw_ops, 0, size, 0);

	return (font);
}

//**********************************************************
/**
	This loads mode specific graphics (mostly fonts).
*/
EVALUE APP::graph_load_by_mode(int mode)
{
	FILE_TextD("DEBUG: Entered LoadGraphByMode().\n");
	
	switch (mode)
	{
	case MSETTINGS:
		/* Settings font. */
		dFont = load_ttf(FONT_S_NAME, FONT_S_SIZE, DMF_A_NAME);
		TTF_SetFontStyle(dFont, FONT_S_STYLE);
		if (!dFont)
		{
			FILE_TextI("ERROR: Couldn't load font (id: %u).\n", FONT_S_NAME);
			throw Exception("Failed to load font!", __LINE__, __FILE__);
		}

		break;

	case MGAME:
		/* Chat font. */
		fChat = load_ttf(CHAT_FONT_NAME, CHAT_FONT_SIZE, DMF_A_NAME);
		TTF_SetFontStyle(fChat, CHAT_FONT_STYLE);
		if (!fChat)
		{
			FILE_TextI("ERROR: Couldn't load font (id: %u).\n", FONT_S_NAME);
			throw Exception("Failed to load font!", __LINE__, __FILE__);
		}

		/* Set TEXTBOX font and color. */
		g.set_chat(&cChat, fChat);

		/* Set input. */
		input_font = fChat;
		input_color = MLText;
		input_w = BG_WCHATW;
		input_h = BG_WCHATH;
		input_x = BG_WCHATX;
		input_y = BG_WCHATY;

		/* Font for buttons. */
		dFont = load_ttf(FONT_S_NAME, FONT_S_SIZE, DMF_A_NAME);
		TTF_SetFontStyle(dFont, FONT_S_STYLE);
		if (!dFont)
		{
			FILE_TextI("ERROR: Couldn't load font (id: %u).\n", FONT_S_NAME);
			throw Exception("Failed to load font!", __LINE__, __FILE__);
		}

		break;

	case MNICK:
		/* Font for nicks. */
		dFont = load_ttf(FONT_NJ_NAME, FONT_NJ_SIZE, DMF_A_NAME);
		TTF_SetFontStyle(dFont, FONT_NJ_STYLE);
		if (!dFont)
		{
			FILE_TextI("ERROR: Couldn't load font (id: %u).\n", FONT_S_NAME);
			throw Exception("Failed to load font!", __LINE__, __FILE__);
		}

		/* Set input. */
		input_font = dFont;
		input_color = MLText;
		input_w = BN_W;
		input_h = BN_H;
		input_x = 0; /* This will be set when needed. */
		input_y = 0; /* and this too. */

		break;

	case MJOIN:
		/* Font for games. */
		dFont = load_ttf(FONT_NJ_NAME, FONT_NJ_SIZE, DMF_A_NAME);
		TTF_SetFontStyle(dFont, FONT_NJ_STYLE);
		if (!dFont)
		{
			FILE_TextI("ERROR: Couldn't load font (id: %u).\n", FONT_S_NAME);
			throw Exception("Failed to load font!", __LINE__, __FILE__);
		}
		
		/* Set input. */
		input_font = dFont;
		input_color = MLText;
		input_w = BC_W;
		input_h = BC_H;
		input_x = BC_STATE_X;
		input_y = BC_STATE_Y;

		break;
	}

	FILE_TextD("DEBUG: Exiting LoadGraphByMode().\n");

	return (A_OK);
}

//**********************************************************
/**
	This function releases resources by mode.
*/
EVALUE APP::release_by_mode(int mode)
{
	switch (mode)
	{
	case MSETTINGS:
		if (dFont)
		{
			TTF_CloseFont(dFont);
			dFont = NULL;
		}
		break;

	case MGAME:
	{
		if (fChat)
		{
			TTF_CloseFont(fChat);
			fChat = NULL;
		}

		if (dFont)
		{
			TTF_CloseFont(dFont);
			dFont = NULL;
		}

		/* Quit game. */
		g.quit();

		/* Close network. */
		NET_Close(true);

		if (st.snd_enabled)
		{
			/* Create little fade. */
			for (int i = st.volume; i > 0; i--)
			{
				SND_Volume(i);

				SDL_Delay(5);
			}

			/* Stop playing playlist. */
			play_list.stop();
		}

		break;
	}

	case MNICK:
		if (dFont)
		{
			TTF_CloseFont(dFont);
			dFont = NULL;
		}
		break;

	case MJOIN:
		if (dFont)
		{
			TTF_CloseFont(dFont);
			dFont = NULL;
		}
		break;
	}

	return (A_OK);
}

