/*
	Blokus SDL
		blfirst.cpp

	This file includes resource load and release
	functions for class APP.

	Copyright (c) 2003-2004 Antti Partanen


	This file is part of Blokus SDL.

	Blokus SDL is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Blokus SDL is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with Blokus SDL; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//**********************************************************
//* INCLUDES
#include "dmf.h"
#include "txtfile.h"

#include "blapp.h"
#include "bljoin.h"
#include "blmenu.h"
#include "blnick.h"
#include "blsettings.h"
#include "bltables.h"
#include "blvideo.h"
#include "blokus.h"
#include "blexceptions.h"
#include "blfiles.h"

//**********************************************************
//* METHODS

//**********************************************************
//* class APP

//**********************************************************
/**
	This function resets many things at startup.
*/
EVALUE APP::init(void)
{
	/* Variables. */
	EVALUE err;
	int i;
	int w, h;

	/* DEBUG info to LOG. */
	FILE_TextD("DEBUG: Entered AtStart().\n");
	
	/* Load graphics. */
	err = graph_load();

	/* If failed, throw exception and print error to LOG file. */
	if (err != A_OK)
	{
		FILE_TextI("ERROR: Function LoadGraph() failed "
			       "to load graphics. (line:%i@file:%s)\n",
				   __LINE__, __FILE__);

		throw Exception("Function LoadGraph() failed to load graphics.",
			            __LINE__, __FILE__);
	}

	/* Enable repeat for keyboard. */
	if (SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_DELAY,
		SDL_DEFAULT_REPEAT_INTERVAL) != 0)
	{
		FILE_TextI("WARNING: Strange, failed enabling keyboard repeat.\n");
	}

	/* Enable unicode translation. */
	SDL_EnableUNICODE(1);

	/* Some LOG info. */
	FILE_TextE("EXTRA: Resetting buttons-classes.\n");

	/* Reset buttons. */
	bmen.B.~BUTTONS();
	bset.B.~BUTTONS();
	bnic.B.~BUTTONS();
	bcre.B.~BUTTONS();
	bgam.B.~BUTTONS();

	/* More LOG info. */
	FILE_TextE("EXTRA: Setting buttons-classes.\n");

	/* Main-menu buttons. */
	for (i = 0; i < bMN; i++)
	{
		bmen.state[i] = 0;
		
		TTF_SizeText(MFont, bmenuT[i].name, &w, &h);
		bmen.B.add(bmenuT[i].x - (w >> 1),
			       bmenuT[i].y, w, bmenuT[i].h,
			       bmenuT[i].name);
	}
	bmen.bt = bmenuT;

	/* Settings-menu buttons. */
	for (i = 0; i < bSN; i++)
	{
		bset.state[i] = 0;

		if (i == bSexit || i == bScancel)
		{
			TTF_SizeText(MFont, bsetT[i].name, &w, &h);
			bset.B.add(bsetT[i].x, bsetT[i].y, w,
				       h, bsetT[i].name);
			continue;
		}

		bset.B.add(bsetT[i].x, bsetT[i].y, bsetT[i].w,
			       bsetT[i].h, bsetT[i].name);
	}
	bset.bt = bsetT;
	bset.infos = SETTINGS_INFO;

	/* Nick-menu buttons. */
	for (i = 0; i < bNN; i++)
	{
		bnic.state[i] = 0;
		TTF_SizeText(MFont, bnicT[i].name, &w, &h);
		bnic.B.add(bnicT[i].x, bnicT[i].y, w,
			       h, bnicT[i].name);
	}
	bnic.bt = bnicT;

	/* Nick selection buttons. */
	for (i = bNN; i < (bNN + MAX_NICKS); i++)
	{
		bnic.state[i] = 0;
		bnic.B.add(BN_NICK_X, BN_NICK_Y + BN_NICK_ADD * (i - bNN),
			       BN_W, BN_H, "<NONE>");
	}

	/* Join-menu buttons. */
	for (i = 0; i < bCN; i++)
	{
		bcre.state[i] = 0;

		if (i == bCstate)
		{
			bcre.B.add(bcreT[i].x, bcreT[i].y, bcreT[i].w,
				       bcreT[i].h, bcreT[i].name);
			continue;
		}
		else
		{
			TTF_SizeText(MFont, bcreT[i].name, &w, &h);
		}

		bcre.B.add(bcreT[i].x, bcreT[i].y, w,
			       h, bcreT[i].name);
	}
	bcre.bt = bcreT;

	/* Game selection buttons. */
	for (i = bCN; i < (bCN + MAX_GAMES); i++)
	{
		bcre.state[i] = 0;
		bcre.B.add(BC_X, BC_Y + BC_ADD * (i - bCN),
			       BC_W, BC_H, "<NONE>");
	}

	/* Game-menu buttons. */
	for (i = 0; i < bGN; i++)
	{
		bgam.state[i] = 0;

		bgam.B.add(bgamT[i].x, bgamT[i].y, bgamT[i].w,
			       bgamT[i].h, bgamT[i].name);
	}
	bgam.bt = bgamT;

	/* Some LOG info. */
	FILE_TextE("INFO: Allocating memory for input-buffer.\n");

	/* Allocate memory for input. */
	input = new char[MAX_INPUT + 2];
	
	/* Check allocation failures. */
	if (!input)
	{
		FILE_TextI("ERROR: Out of memory, couldn't allocate"
			       " memory for input-buffer. (line:%i@file:%s)\n",
				   __LINE__, __FILE__);

		throw OutOfMemory("Out of memory, couldn't allocate "
			              "memory for input-buffer.",
						  __LINE__, __FILE__);
	}

	/* Some LOG info. */
	FILE_TextE("EXTRA: Setting volume to default.\n");

	/* Set volume. */
	SND_Volume(st.volume);

	FILE_TextD("DEBUG: Exiting AtStart()-function.\n");

	/* Load sounds, if enabled. */
	if (st.snd_enabled)
	{
		snd_gameover = SND_LoadFromDMF(GAMEOVER, DMF_A_NAME);
		snd_next = SND_LoadFromDMF(NEXT, DMF_A_NAME);
	}
	/* Sound disabled, set IDs to none. */
	else
	{
		snd_gameover = SND_ID_NONE;
		snd_next = SND_ID_NONE;
	}

	/* Load music IDs, if music DMF exist. */
	int lx;
	int ly = 0;
	DMFHEAD dmfhead;

	if (!st.snd_enabled)
	{
		FILE_TextD("DEBUG: Music not loaded, becose "
			       "sounds are disabled.\n");
	}
	else if (DMF_ReadHead(&dmfhead, DMF_MUSIC) != 0)
	{
		FILE_TextI("INFO: Music not loaded, becose "
			       "DMF_ReadHead() failed to open DMF: %s\n",
			       DMF_MUSIC);
	}
	else
	{
		for (lx = 0; ; lx++)
		{
			DMFINFO dmfinfo;

			if (DMF_ReadInfo(lx, &dmfinfo, DMF_MUSIC) != A_OK)
			{
				break;
			}

			if (dmfinfo.title[0] != '#')
			{
				continue;
			}

			if (!play_list.add(lx))
			{
				break;
			}

			FILE_TextE("EXTRA: Added song from DMF "
				       "\"%s\" with id %i and name "
					   "\"%s\" to playlist.\n",
					   DMF_MUSIC, lx, dmfinfo.title);

			ly++;

			if (ly >= MAX_SONGS)
			{
				break;
			}
		}

		if (ly > 0)
		{
			play_list.loop(true);
			play_list.random(true);
		}
	}

	/* Return. */
	return (A_OK);
}

//**********************************************************
/**
	Here should be released all resources
	and this function should be called
	before closing the application.
*/
void APP::release(void)
{
	/* Info for debug LOG. */
	FILE_TextD("DEBUG: Entered Release()-function.\n");

	if (input)
	{
		delete[] input;
		input = NULL;
	}
	
	graph_release();

	NET_Quit();

	TTF_Quit();
	VID_Quit();
	SND_Quit();
	SDL_Quit();

	/* Info for debug LOG. */
	FILE_TextD("DEBUG: Exiting Release()-function.\n");

	/* Return. */
	return;
}


