/*
	Blokus SDL
		blsound.cpp

	This file includes sound handling functions
	to blokus.

	Copyright (c) 2003-2004 Antti Partanen


	This file is part of Blokus SDL.

	Blokus SDL is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Blokus SDL is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with Blokus SDL; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

//**********************************************************
//* INCLUDES
#include <stdlib.h>
#include <string.h>

#include "SDL/SDL.h"
#include "SDL/SDL_thread.h"
#include "SDL/SDL_sound.h"
#include "txtfile.h"
#include "dmf.h"
#include "derrors.h"

#include "blsound.h"

//**********************************************************
//* GLOBAL VARIABLES
Sound_Sample *SND_Samples[MAX_SOUND_SAMPLES];
SDL_RWops *SND_Samples_RWops[MAX_SOUND_SAMPLES];
char *SND_Samples_data[MAX_SOUND_SAMPLES];
bool SND_Samples_playing[MAX_SOUND_SAMPLES];
SND_PLAYLIST *SND_Samples_playlist[MAX_SOUND_SAMPLES];
static Uint8 *decoded_ptr[MAX_SOUND_SAMPLES];
static Uint32 decoded_bytes[MAX_SOUND_SAMPLES];

int volume = 0;

//**********************************************************
//* FUNCTION DEFINITIONS
static int read_more_data(int id);
static void audio_callback(void *userdata, Uint8 *stream, int len);
int play_list(void *data);

//**********************************************************
//* FUNCTIONS

//**********************************************************
/**
	Initializes sounds.
*/
EVALUE SND_Init()
{
	if (!Sound_Init())
	{
		return (E_FAILED);
	}

	for (int i = 0; i < MAX_SOUND_SAMPLES; i++)
	{
		SND_Samples[i] = NULL;
		SND_Samples_RWops[i] = NULL;
		SND_Samples_data[i] = NULL;
		SND_Samples_playing[i] = false;
		SND_Samples_playlist[i] = NULL;
		decoded_ptr[i] = NULL;
		decoded_bytes[i] = 0;
	}

	//* Set up audio specs.
	SDL_AudioSpec specs;
	specs.freq = 44100;
	specs.format = AUDIO_S16SYS;
	specs.channels = 2;
	specs.samples = SOUND_AUDIO_BUFFER_BYTES;
	specs.callback = (void (*)(void*,Uint8*,int))audio_callback;
	specs.userdata = NULL;

	//* Now try to open the audio device.
	if (SDL_OpenAudio(&specs, NULL) < 0)
	{
		return (E_FAILED);
	}

	SDL_PauseAudio(0);

	//* For now the audio_callback-thread will handle the playing.
	return (A_OK);
}

/**********************************************************/
/**
	Deinitializes sounds.
*/
void SND_Quit()
{
	SND_StopAll();

	for (int i = 0; i < MAX_SOUND_SAMPLES; i++)
	{
		SND_Release(i);
	}

	Sound_Quit();
}

/**********************************************************/
/**
	Load sound from DMF to buffer.
	
	@return
	 Returns buffer ID.
*/
int SND_LoadFromDMF(int id, char *dmf)
{
	int i;
	int snd_id = SND_ID_NONE;

	//* Find free buffer for the sound.
	for (i = 0; i < MAX_SOUND_SAMPLES; i++)
	{
		if (!SND_Samples[i])
		{
			snd_id = i;
			break;
		}
	}

	if (snd_id == SND_ID_NONE)
	{
		FILE_TextI("ERROR: No free sound buffer found.\n");
		return (SND_ID_NONE);
	}

	//* Read info from DMF and try to find and load
	//* wanted data.
	DMFINFO dmfinfo;
	DMFHEAD dmfhead;

	if (DMF_ReadHead(&dmfhead, dmf) != 0)
	{
		FILE_TextI("ERROR: DMF_ReadHead() failed to open DMF: %s\n",
			       dmf);
		return (SND_ID_NONE);
	}
 
	if (DMF_ReadInfo(id, &dmfinfo, dmf) != A_OK)
	{
		FILE_TextI("ERROR: DMF_ReadInfo() failed to read"
			       " info for sound. (id: %u)\n", id);
		return (SND_ID_NONE);
	}
	
	int data_size;
	char *data = DMF_DataRead(id, &data_size, dmf);

	if (!data)
	{
		FILE_TextI("ERROR: Unable to load sound data from DMF."
			       "(id: %u)\n", id);
		return (SND_ID_NONE);
	}

	//* Data loaded, set up the sound to buffer.
	SDL_RWops *rw_ops = SDL_RWFromMem(data, data_size);

	if (!rw_ops)
	{
		FILE_TextI("ERROR: Unable to create SDL_RWops for sound buffer.\n");
		DMF_Free(data);
		return (SND_ID_NONE);
	}

	Sound_Sample *sample = Sound_NewSample(rw_ops, NULL, NULL,
		                                   SOUND_DECODE_BUFFER_BYTES);

	
	if (!sample)
	{
		FILE_TextI("ERROR: Unable to create new sample"
			       "to sound buffer.\n");
		SDL_RWclose(rw_ops);
		DMF_Free(data);
		return (SND_ID_NONE);
	}
	
	//* Set up pointers.
	SDL_LockAudio();
	SND_Samples[snd_id] = sample;
	SND_Samples_RWops[snd_id] = rw_ops;
	SND_Samples_data[snd_id] = data;
	SND_Samples_playing[snd_id] = false;
	SND_Samples_playlist[i] = NULL;
	decoded_ptr[snd_id] = NULL;
	decoded_bytes[snd_id] = 0;
	SDL_UnlockAudio();

	return (snd_id);
}

//**********************************************************
/**
	Deinitializes sounds.
*/
void SND_Play(int id)
{
	if (id >= MAX_SOUND_SAMPLES)
	{
		return;
	}

	if (!SND_Samples[id] || SND_Samples_playing[id])
	{
		return;
	}

	SDL_LockAudio();
	SND_Samples_playing[id] = true;
	SDL_UnlockAudio();
}

//**********************************************************
/**
	Stops playing a sound.
*/
void SND_Stop(int id)
{
	if (id >= MAX_SOUND_SAMPLES)
	{
		return;
	}

	if (!SND_Samples[id] || !SND_Samples_playing[id])
	{
		return;
	}

	SDL_LockAudio();
	SND_Samples_playing[id] = false;
	decoded_ptr[id] = NULL;
	decoded_bytes[id] = 0;
	SDL_UnlockAudio();
}

//**********************************************************
/**
	Stops playing a sound.
*/
void SND_StopAll(void)
{
	SDL_LockAudio();
	for (int i= 0; i < MAX_SOUND_SAMPLES; i++)
	{
		SND_Samples_playing[i] = false;
	}
	SDL_UnlockAudio();
}

//**********************************************************
/**
	Set volume of all sound to given (0-100).
*/
void SND_Volume(int _volume)
{
	if (_volume > 100 || _volume < 0)
	{
		return;
	}

	SDL_LockAudio();
	volume = _volume * SDL_MIX_MAXVOLUME / 100;
	SDL_UnlockAudio();
}

//**********************************************************
/**
	Pauses avery audio playback.
*/
void SND_Pause(int pause)
{
	SDL_LockAudio();
	SDL_PauseAudio(pause);
	SDL_UnlockAudio();
}

//**********************************************************
/**
	Pauses every audio playback.
*/
void SND_Release(int id)
{
	if (id >= MAX_SOUND_SAMPLES)
	{
		return;
	}

	SDL_LockAudio();

	if (SND_Samples[id])
	{
		SND_Samples_playing[id] = false;

//		SDL_RWclose(SND_Samples_RWops[id]);
//		SND_Samples_RWops[id] = NULL;
		DMF_Free(SND_Samples_data[id]);
		SND_Samples_data[id] = NULL;
		Sound_FreeSample(SND_Samples[id]);
		SND_Samples[id] = NULL;
		decoded_ptr[id] = NULL;
		decoded_bytes[id] = 0;
	}

	SDL_UnlockAudio();
}

//**********************************************************
/**
	This updates (decoded_bytes) and (decoder_ptr) with more audio data,
	taking into account looping and/or predecoding.
*/
static int read_more_data(int id)
{
	if (!SND_Samples_playing[id])
	{
		return (0);
	}

	Sound_Sample *sample = SND_Samples[id];
	
	//* don't need more data; just return.
	if (decoded_bytes[id] > SOUND_BUFFER_MINIMUM)
	{
		return (decoded_bytes[id]);
	}

	//* need more. See if there's more to be read...
	if (!(sample->flags & (SOUND_SAMPLEFLAG_ERROR | SOUND_SAMPLEFLAG_EOF)))
	{
		decoded_bytes[id] += Sound_Decode(sample);
		decoded_ptr[id] = (Uint8 *)sample->buffer;
		return (read_more_data(id));
	}

	//* No more to be read from stream.

	return(0);
}

//**********************************************************
/**
	Load more data into sound buffer, when needed.
*/
static void audio_callback(void *userdata, Uint8 *stream, int len)
{
	//* Clear the stream.
	memset(stream, '\0', len);

	//* Start mixing playing sounds to the stream.
	for (int i = 0; i < MAX_SOUND_SAMPLES; i++)
	{
		if (!SND_Samples[i])
		{
			continue;
		}

		if (!SND_Samples_playing[i])
		{
			continue;
		}

		//* Current sample.
		Sound_Sample *sample = SND_Samples[i];
		
		//* bytes written to stream this time through the callback.
		int bw = 0;

		while (bw < len)
		{
			//* bytes to copy on this iteration of the loop.
			unsigned int cpysize = len - bw;

			if (cpysize > decoded_bytes[i])
			{
				cpysize = decoded_bytes[i];
			}

			if (cpysize > 0)
			{
				SDL_MixAudio(stream, decoded_ptr[i], cpysize, volume);
				bw += cpysize;
				decoded_ptr[i] += cpysize;
				decoded_bytes[i] -= cpysize;
			}

			//* read more data, if needed.
			//* decoded_bytes and decoder_ptr are updated as necessary...
			if (!read_more_data(i))
			{
				//* ...there isn't any more data to read!
				SND_Samples_playing[i] = false;
				
				if (SND_Samples_playlist[i])
				{
					SND_Samples_playlist[i]->next();
				}

				return;
			}
		}
	}
}

//**********************************************************
/*
	class PLAYLIST.
*/

//**********************************************************
/**
	Constructor.
*/
SND_PLAYLIST::SND_PLAYLIST()
{
	/* Reset everything. */
	for (int i = 0; i < SND_LIST_MAX; i++)
	{
		ids[i] = SND_LIST_NONE;
	}

	vrandom = false;
	vloop = false;
	vplay = false;
	vplaying = false;
	current = 0;
	nof = 0;

	semaphore = NULL;
}

//**********************************************************
/**
	Destructor.
*/
SND_PLAYLIST::~SND_PLAYLIST()
{
	/* Try to shutdown the play_list-thread nicely. */
	if (!vplaying)
	{
		return;
	}

	if (vplaying)
	{
		SND_Stop(current);
	}

	vplay = false;
	vplaying = false;

	/* Reset everything. */
	if (semaphore)
	{
		SDL_DestroySemaphore(semaphore);
	}

	SND_PLAYLIST();
}

//**********************************************************
/**
	Pack samples to beginning of the list.
*/
void SND_PLAYLIST::pack(void)
{
	int empty = 0xffff;

	for (int i = 0; i < SND_LIST_MAX; i++)
	{
		if (ids[i] == SND_LIST_NONE && i < empty)
		{
			empty = i;
		}
		else if (i > empty)
		{
			ids[empty] = ids[i];
			ids[i] = SND_LIST_NONE;
			empty++;
		}
	}
}

//**********************************************************
/**
	Add new item to list.
*/
bool SND_PLAYLIST::add(int id)
{
	if (semaphore)
	{
		SDL_SemWait(semaphore);
	}

	/* Find empty place from ids-list. */
	for (int i = 0; i < SND_LIST_MAX; i++)
	{
		if (ids[i] == SND_LIST_NONE)
		{
			ids[i] = id;
			nof++;
			pack();

			if (semaphore)
			{
				SDL_SemPost(semaphore);
			}

			return (true);
		}
	}

	if (semaphore)
	{
		SDL_SemPost(semaphore);
	}

	return (false);
}

//**********************************************************
/**
	Remove item from list.
*/
bool SND_PLAYLIST::remove(int id)
{
	if (semaphore)
	{
		SDL_SemWait(semaphore);
	}

	for (int i = 0; i < SND_LIST_MAX; i++)
	{
		if (ids[i] == id)
		{
			ids[i] = SND_LIST_NONE;
			nof--;
			pack();

			if (semaphore)
			{
				SDL_SemPost(semaphore);
			}

			return (true);
		}
	}

	if (semaphore)
	{
		SDL_SemPost(semaphore);
	}

	return (false);
}

//**********************************************************
/**
	Release all and remove all items within list.
*/
void SND_PLAYLIST::release_all(void)
{
	stop();

	if (semaphore)
	{
		SDL_SemWait(semaphore);
	}

	if (current != SND_ID_NONE)
	{
		SND_Release(current);
	}

	for (int i = 0; i < SND_LIST_MAX; i++)
	{
		if (ids[i] != SND_LIST_NONE)
		{
			ids[i] = SND_LIST_NONE;
		}
	}

	nof = 0;

	if (semaphore)
	{
		SDL_SemPost(semaphore);
	}
}

//**********************************************************
/**
	Set list as randomized.
*/
void SND_PLAYLIST::random(bool r)
{
	if (semaphore)
	{
		SDL_SemWait(semaphore);
	}

	vrandom = r;

	if (semaphore)
	{
		SDL_SemPost(semaphore);
	}
}

//**********************************************************
/**
	Enable looping.
*/
void SND_PLAYLIST::loop(bool l)
{
	if (semaphore)
	{
		SDL_SemWait(semaphore);
	}

	vloop = l;

	if (semaphore)
	{
		SDL_SemPost(semaphore);
	}
}

//**********************************************************
/**
	Play this list.
*/
EVALUE SND_PLAYLIST::play(char *name)
{
	/* Do some basic checks. */
	if (vplaying)
	{
		return (A_OK);
	}

	/* Create semaphore. */
	if (!semaphore)
	{
		semaphore = SDL_CreateSemaphore(1);
	}

	if (!semaphore)
	{
		return (E_NULL);
	}

	/* Set on. */
	vplay = true;
	vplaying = false;

	/* Set name. */
	strcpy(dmf, name);

	/* Create play_list-thread. */
	next();

	return (A_OK);
}

//**********************************************************
/**
	Stop playing.
*/
void SND_PLAYLIST::stop(void)
{
	if (semaphore)
	{
		SDL_SemWait(semaphore);
	}

	if (vplaying)
	{
		SND_Stop(current);
	}

	vplay = false;
	vplaying = false;

	if (semaphore)
	{
		SDL_SemPost(semaphore);
	}
}

//**********************************************************
/**
	Play next item in the list.
*/
void SND_PLAYLIST::next(void)
{
	static unsigned int played;
	static int next;

	if (semaphore)
	{
		SDL_SemWait(semaphore);
	}

	if (!vplay)
	{
		if (semaphore)
		{
			SDL_SemPost(semaphore);
		}

		return;
	}

	if (!vplaying)
	{
		played = 0;
		next = -1;
	}
	else
	{
		SND_Stop(current);
		SND_Release(current);

		played++;

		if (!vloop && played >= nof)
		{
			vplay = false;
			vplaying = false;

			if (semaphore)
			{
				SDL_SemPost(semaphore);
			}
			
			return;
		}
	}

	if (vrandom)
	{
		next = rand() * nof / RAND_MAX;
	}
	else
	{
		next++;
	}
	
	if (ids[next] == SND_LIST_NONE)
	{
		next = 0;
	}

	if (ids[next] == SND_LIST_NONE)
	{
		if (semaphore)
		{
			SDL_SemPost(semaphore);
		}

		return;
	}

	current = SND_LoadFromDMF(ids[next], dmf);

	if (current == SND_ID_NONE)
	{
		if (semaphore)
		{
			SDL_SemPost(semaphore);
		}

		return;
	}

	SND_Play(current);

	vplaying = true;

	if (semaphore)
	{
		SDL_SemPost(semaphore);
	}
}

