/*
	Blokus SDL
		blsound.h

	This file includes sound handling functions
	to blokus.

	Copyright (c) 2003-2004 Antti Partanen


	This file is part of Blokus SDL.

	Blokus SDL is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Blokus SDL is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with Blokus SDL; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef BLSOUND_H
#define BLSOUND_H

//**********************************************************
//* INCLUDES
#include "SDL/SDL_thread.h"
#include "derrors.h"

//**********************************************************
//* DEFINES

/**
	Defines how many sounds can be loaded simultaneously.
*/
#define MAX_SOUND_SAMPLES 8

/**
	Defines the sound decoding buffer size.
*/
#define SOUND_DECODE_BUFFER_BYTES 0x10000 //* 64kB

/**
	Defines the audio buffer size, where data is decoded.
*/
#define SOUND_AUDIO_BUFFER_BYTES 0x2000 //* 8kB

/**
	Defines the minimum amount of data in audio buffer.
*/
#define SOUND_BUFFER_MINIMUM 1024 //* 1kB

/**
	Defines none ID sound.
*/
#define SND_ID_NONE 0xffff

/**
	Defines maximum for playlist items.
*/
#define SND_LIST_MAX 12

/**
	Defines none ID for playlist.
*/
#define SND_LIST_NONE 0xffff

/**
	Defines delay for playlist.
*/
#define SND_LIST_DELAY 70

/**
	Defines class PLAYLIST, which is used for playing music.
*/
class SND_PLAYLIST
{
private:
	//*********************************
	/* Private variables. */
	
	/** Name of DMF, where to load songs when playing them. */
	char dmf[320];

	//*********************************
	/* Private functions. */
	void pack(void);

public:
	//*********************************
	/* Public methods. */
	/* Constructor and destructor. */
	SND_PLAYLIST();
	~SND_PLAYLIST();

	/* Add item to playlist. */
	bool add(int id);

	/* Remove item from playlist. */
	bool remove(int id);

	/* Release everything in playlist. */
	void release_all();

	/* Randomize playlist. */
	void random(bool random);

	/* Loop playlist. */
	void loop(bool loop);

	/* Play playlist. */
	EVALUE play(char *dmf);

	/* Stop playing. */
	void stop(void);

	/* Next song. */
	void next(void);

	//*********************************
	/* Public Variables. */
	
	/** Whether to play randomized or not. */
	bool vrandom;

	/** Whether to play as looping or not. */
	bool vloop;

	/** Playing or not. */
	bool vplaying;

	/** Play/stop. */
	bool vplay;

	/** Number of items in list. */
	unsigned int nof;

	/** List of item IDs in DMF. */
	int ids[SND_LIST_MAX];

	/** Currently playing item. */
	int current;

	/** Semaphore for using item list. */
	SDL_sem *semaphore;
};

//**********************************************************
//* Functions

/** Initialize sounds. */
EVALUE SND_Init(void);

/** Deinitialize sounds. */
void SND_Quit(void);

/** Stop playing sound with selected ID. */
void SND_Release(int id);

/** Get sound from DMF. */
int SND_LoadFromDMF(int id, char *dmf);

/** Play sound with selected ID. */
void SND_Play(int id);

/** Stop playing sound with selected ID. */
void SND_Stop(int id);

/** Set volume of all sound to given (0-100). */
void SND_Volume(int volume);

/** Pauses every audio playback. */
void SND_Pause(int pause);

/** Stops playing a sound. */
void SND_StopAll(void);

/** Starts playing given playlist. */
void SND_PlayList(SND_PLAYLIST *list);

#endif

